local LHR = select(2, ...)

local L = LHR.L

-- Adds healer to global table and one of the two rotation tables
function LHR:registerHealer(healerName)

    -- Initialize healer 'object'
    local healer = {}
    healer.name = healerName
    healer.GUID = UnitGUID(healerName)
    healer.frame = nil
    healer.nextHeal = false
    healer.lastHealTime = 0

    -- Add to global list
    table.insert(LHR.healerTable, healer)

    -- Add to rotation or backup group depending on rotation group size
    if (#LHR.rotationTables.rotation > 2) then
        table.insert(LHR.rotationTables.backup, healer)
    else
        table.insert(LHR.rotationTables.rotation, healer)
    end

    LHR:drawHealerFrames()

    return healer
end

-- Removes a healer from all lists
function LHR:removeHealer(deletedHealer)

    -- Clear from global list
    for key, healer in pairs(LHR.healerTable) do
        if (healer.name == deletedHealer.name) then
            LHR:hideHealer(healer)
            table.remove(LHR.healerTable, key)
            break
        end
    end

    -- clear from rotation lists
    for key, healerTable in pairs(LHR.rotationTables) do
        for subkey, healer in pairs(healerTable) do
            if (healer.name == deletedHealer.name) then
                table.remove(healerTable, subkey)
            end
        end
    end

    LHR:drawHealerFrames()
end

-- Update the rotation list once a heal has been done.
-- The parameter is the healer that used it's heal (successfully or not)
function LHR:rotate(lastHealer, rotateWithoutCooldown)

    local playerName, realm = UnitName("player")
    local lastHealerRotationTable = LHR:getHealerRotationTable(lastHealer)
    lastHealer.lastHealTime = GetTime()

    -- Do not trigger cooldown when rotation from a dead or disconnected status
    if (rotateWithoutCooldown ~= true) then
        LHR:startHealerCooldown(lastHealer)
    end

    local nextHealer = nil

    if (lastHealerRotationTable == LHR.rotationTables.rotation) then
        nextHealer = LHR:getNextRotationHealer(lastHealer)

        if (nextHealer ~= nil) then
            LHR:setNextHeal(nextHealer)
        end
    end
end

-- Removes all nextHeal flags and set it true for next shooter
function LHR:setNextHeal(nextHealer)
    for key, healer in pairs(LHR.rotationTables.rotation) do
        if (healer.name == nextHealer.name) then
            healer.nextHeal = true

            if (nextHealer.name == UnitName("player")) and LHR.db.profile.enableNextToHealSound then
                PlaySoundFile(LHR.constants.sounds.nextToHeal)
            end
        else
            healer.nextHeal = false
        end

        LHR:refreshHealerFrame(healer)
    end
end

-- Check if the player is the next in position to heal
function LHR:isPlayerNextHeal()

    if(not LHR:isHealer("player")) then
        return false
    end

    local player = LHR:getHealer(nil, UnitGUID("player"))

    if (not player.nextHeal) then

        local isRotationInitialized = false;
        local rotationTable = LHR.rotationTables.rotation

        -- checking if a healer is flagged nextHeal
        for key, healer in pairs(rotationTable) do
            if (healer.nextHeal) then
                isRotationInitialized = true;
                break
            end
        end

        -- First in rotation has to heal if not one is flagged
        if (not isRotationInitialized and LHR:getHealerIndex(player, rotationTable) == 1) then
            return true
        end

    end

    return player.nextHeal
end

-- Find and returns the next healer that will heal base on last shooter
function LHR:getNextRotationHealer(lastHealer)

    local rotationTable = LHR.rotationTables.rotation
    local nextHealer
    local lastHealerIndex = 1

    -- Finding last healer index in rotation
    for key, healer in pairs(rotationTable) do
        if (healer.name == lastHealer.name) then
            lastHealerIndex = key
            break
        end
    end

    -- Search from last healer index if not last on rotation
    if (lastHealerIndex < #rotationTable) then
        for index = lastHealerIndex + 1 , #rotationTable, 1 do
            local healer = rotationTable[index]
            if (LHR:isEligibleForNextHeal(healer)) then
                nextHealer = healer
                break
            end
        end
    end

    -- Restart search from first index
    if (nextHealer == nil) then
        for index = 1 , lastHealerIndex, 1 do
            local healer = rotationTable[index]
            if (LHR:isEligibleForNextHeal(healer)) then
                nextHealer = healer
                break
            end
        end
    end

    -- If no healer in the rotation match the alive/online/CD criteria
    -- Pick the healer with the lowest cooldown
    if (nextHealer == nil and #rotationTable > 0) then
        local latestHeal = GetTime() + 1
        for key, healer in pairs(rotationTable) do
            if (LHR:isHealerAliveAndOnline(healer) and healer.lastHealTime < latestHeal) then
                nextHealer = healer
                latestHeal = healer.lastHealTime
            end
        end
    end

    return nextHealer
end

-- Init/Reset rotation status, next heal is the first healer on the list
function LHR:resetRotation()
    for key, healer in pairs(LHR.rotationTables.rotation) do
        healer.nextHeal = false
        LHR:refreshHealerFrame(healer)
    end
end

-- @todo: remove this | TEST FUNCTION - Manually rotate healers for test purpose
function LHR:testRotation()

    for key, healer in pairs(LHR.rotationTables.rotation) do
        if (healer.nextHeal) then
            LHR:rotate(healer, false)
            break
        end
    end
end

-- Check if a healer is already registered
function LHR:isHealerRegistered(GUID)

    -- @todo refactor this using LHR:getHealer(name, GUID)
    for key,healer in pairs(LHR.healerTable) do
        if (healer.GUID == GUID) then
            return true
        end
    end

    return false
end

-- Return our healer object from name or GUID
function LHR:getHealer(name, GUID)

    for key,healer in pairs(LHR.healerTable) do
        if ((GUID ~= nil and healer.GUID == GUID) or (name ~= nil and healer.name == name)) then
            return healer
        end
    end

    return nil
end

function LHR:getHealerClass(name, GUID)

    for key,healer in pairs(LHR.healerTable) do
        if ((GUID ~= nil and healer.GUID == GUID) or (name ~= nil and healer.name == name)) then
            return class_name
        end
    end

    return nil
end

-- Iterate over healer list and purge healer that aren't in the group anymore
function LHR:purgeHealerList()
    for _, healer in pairs(LHR.healerTable) do
        if (not UnitInParty(healer.name)) then
            LHR:unregisterUnitEvents(healer)
            LHR:removeHealer(healer)
        end
    end
end

-- Iterate over all raid members to find healers and update their status
function LHR:updateRaidStatus()

    if (LHR:isInPveRaid()) then

        local playerCount = GetNumGroupMembers()

        for index = 1, playerCount, 1 do

            local name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(index)

            -- Players name might be nil at loading
            if (name ~= nil) then
                local GUID = UnitGUID(name)
                local healer

                if(LHR:isHealer(name)) then

                    local registered = LHR:isHealerRegistered(GUID)

                    if (not registered) then
                        if (not InCombatLockdown()) then
                            healer = LHR:registerHealer(name)
                            LHR:registerUnitEvents(healer)
                            registered = true
                        end
                    else
                        healer = LHR:getHealer(nil, GUID)
                    end

                    if (registered) then
                        LHR:updateHealerStatus(healer)
                    end
                end

            end
        end

        LHR:updateDragAndDrop()

        if (not LHR.raidInitialized) then
            if (not LHR.db.profile.doNotShowWindowOnRaidJoin) then
                LHR:updateDisplay()
            end
            LHR:sendSyncOrderRequest()
            LHR.raidInitialized = true
        end
    else
        if(LHR.raidInitialized == true) then
            LHR:updateDisplay()
            LHR.raidInitialized = false
        end
    end

    LHR:purgeHealerList()
end

-- Update healer status
function LHR:updateHealerStatus(healer)

    -- Jump to the next healer if the current one is dead or offline
    if (healer.nextHeal and (not LHR:isHealerAliveAndOnline(healer))) then
        LHR:rotate(healer, false, true)
    end

    LHR:refreshHealerFrame(healer)
end

-- Moves given healer to the given position in the given group (ROTATION or BACKUP)
function LHR:moveHealer(healer, group, position)

    local originTable = LHR:getHealerRotationTable(healer)
    local originIndex = LHR:getHealerIndex(healer, originTable)

    local destinationTable = LHR.rotationTables.rotation
    local finalIndex = position

    if (group == 'BACKUP') then
        destinationTable = LHR.rotationTables.backup
        -- Remove nextHeal flag when moved to backup
        healer.nextHeal = false
    end

    -- Setting originalIndex
    local sameTableMove = originTable == destinationTable

    -- Defining finalIndex
    if (sameTableMove) then
        if (position > #destinationTable or position == 0) then
            if (#destinationTable > 0) then
                finalIndex = #destinationTable
            else
                finalIndex = 1
            end
        end
    else
        if (position > #destinationTable + 1 or position == 0) then
            if (#destinationTable > 0) then
                finalIndex = #destinationTable  + 1
            else
                finalIndex = 1
            end
        end
    end

    if (sameTableMove) then
        if (originIndex ~= finalIndex) then
            table.remove(originTable, originIndex)
            table.insert(originTable, finalIndex, healer)
        end
    else
        table.remove(originTable, originIndex)
        table.insert(destinationTable, finalIndex, healer)
    end

    LHR:drawHealerFrames()
end

-- Find the table that contains given healer (rotation or backup)
function LHR:getHealerRotationTable(healer)
    if (LHR:tableContains(LHR.rotationTables.rotation, healer)) then
        return LHR.rotationTables.rotation
    end
    if (LHR:tableContains(LHR.rotationTables.backup, healer)) then
        return LHR.rotationTables.backup
    end
end

-- Returns a healer's index in the given table
function LHR:getHealerIndex(healer, table)
    local originIndex = 0

    for key, loopHealer in pairs(table) do
        if (healer.name == loopHealer.name) then
            originIndex = key
            break
        end
    end

    return originIndex
end

-- Builds simple rotation tables containing only healers names
function LHR:getSimpleRotationTables()

    local simpleTables = { rotation = {}, backup = {} }

    for key, rotationTable in pairs(LHR.rotationTables) do
        for _, healer in pairs(rotationTable) do
            table.insert(simpleTables[key], healer.name)
        end
    end

    return simpleTables
end

-- Apply a simple rotation configuration
function LHR:applyRotationConfiguration(rotationsTables)

    for key, rotationTable in pairs(rotationsTables) do

        local group = 'ROTATION'
        if (key == 'backup') then
            group = 'BACKUP'
        end

        for index, healerName in pairs(rotationTable) do
            local healer = LHR:getHealer(healerName)
            if (healer) then
                LHR:moveHealer(healer, group, index)
            end
        end
    end
end

-- Display an alert and play a sound when the player should immediatly heal
function LHR:throwHealAlert()
    RaidNotice_AddMessage(RaidWarningFrame, L['HEAL_NOW_LOCAL_ALERT_MESSAGE'], ChatTypeInfo["RAID_WARNING"])

    if (LHR.db.profile.enableHealNowSound) then
        PlaySoundFile(LHR.constants.sounds.alarms[LHR.db.profile.healNowSound])
    end
end

-- Send a defined message to backup player or next rotation player if there's no backup
function LHR:alertBackup(message, nextHealer, noComms)
    local playerName = UnitName('player')
    local player = LHR:getHealer(playerName)

    -- Non healer have no reason to ask for backup
    if (not LHR:isHealer('player')) then
        return
    end

    if (#LHR.rotationTables.backup < 1) then

        if (nextHealer == nil) then
            nextHealer = LHR:getNextRotationHealer(player)
        end

        SendChatMessage(message, 'WHISPER', nil, nextHealer.name)
        if (noComms ~= true) then
            LHR:sendBackupRequest(nextHealer.name)
        end
    else
        LHR:whisperBackup(message)
    end
end

-- Whisper provided message of fail message to all backup except player
function LHR:whisperBackup(message, noComms)

    local name = UnitName("player")

    if (message == nil) then
        message = LHR.db.profile.whisperFailMessage
    end

    for key, backupHealer in pairs(LHR.rotationTables.backup) do
        if (backupHealer.name ~= name) then
            SendChatMessage(message, 'WHISPER', nil, backupHealer.name)

            if (noComms ~= true) then
                LHR:sendBackupRequest(backupHealer.name)
            end
        end
    end
end
